/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file artsstream.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// this header
#include "artsstream.hh"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"

using namespace PIAVE;

PIAVE_PLUGIN( ArtsPlugin );


/*
 *    wrapper plugin
 */
ArtsPlugin::ArtsPlugin() :
  Plugin( "ARTSStream", PLUGIN_OUTSTREAM )
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * monitor = caps->addDaughter( "monitor" );
    PropertyNode * audio = monitor->addDaughter( "audio" );
    audio->addAttr("type","arts");

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Simple ARTS layer for audio output.\n" );

    setCapabilities( caps );

    /* this interface is prioritized over e.g. oss */
    setPriority( 5 );

}


PluginBase * 
ArtsPlugin::newElement( const PropertyNode * p )
{
    return new OutARTSStream();
}






/*
 *  \class OutARTSstream
 *
 *
 */
OutARTSStream::OutARTSStream() : 
  OutAudioStreamIFace(),
  _arts_stream( 0 ),
  _streamName( "piave" )
{
    /* init aRts */
    int err;
    err = arts_init();
    if ( err != 0 ) {
        WARN( "failed to initialize aRts: " << arts_error_text( err)  );
        setState( state_error );
        throw audio_stream_error("failed to initialize aRts");
    }

}

OutARTSStream::~OutARTSStream()
{
    if ( _arts_stream ) {
        arts_close_stream( _arts_stream );
    }
    arts_free();
}


void
OutARTSStream::init( Frame & f ) throw(exception)
{
    if ( getState() == state_error ) return;

    int format = AFMT_S16_NE;
    int frequency_request = -1;
    int channels_request = -1;
    int bitspersample_request = -1;

    InAVStreamIFace * streamContext = 0;
    streamContext = f.getAudioContext();
    if ( !streamContext ) 
    {
        streamContext = f.getAVContext();
        if ( !streamContext ) 
        {
            WARN( "Couldn't get properties of audio codec since decode context is missing." );
            setState( state_error );
            return;
        }
    }

    double frequency = -1.0;
    double channels = -1.0;
    double bitspersample = -1.0;
    PropertyNode * av_props = streamContext->getProperties();
    PropertyNode * audio = 0;
    PropertyNode * p = 0;
    int d = 0;
    string type;
    while( ( p = av_props->getDaughter( "codec", d++, true ) ) && !audio ) {
        if ( p->getAttr( "type", type ) ) {
            if ( type == "audio" ) {
                audio = p;
            }
        }
    }
    if ( !audio ) {
        WARN( "Couldn't get properties of audio codec" );
        setState( state_error );
        return;
    }
    audio->getAttr( "frequency", frequency );
    audio->getAttr( "channels", channels );
    audio->getAttr( "bitspersample", bitspersample );
    frequency_request = static_cast<int>( frequency );
    channels_request = static_cast<int>( channels );
    bitspersample_request = static_cast<int>( bitspersample );
    PDELETE( av_props );

    if ( frequency_request < 0  ||
         channels_request  < 0  || 
         bitspersample_request  < 0  )
    {
        WARN( "couldn't get all parameters"
              <<PRINTV(frequency_request)<<PRINTV(channels_request)<<PRINTV(bitspersample_request) );
        setState( state_error );
        return;
    }


    /* open a play stream */
    _arts_stream = arts_play_stream( frequency_request, bitspersample_request, 
                                     channels_request, _streamName.c_str() );
    if ( _arts_stream == 0 ) {
        WARN( "couldn't open arts stream " );
        setState( state_error );
        return;
    }

    /* do I need to check if I got what I asked for ??? */
    _nChannels = channels_request;
    _frequency = frequency_request;
    
    /* check what we got */
    INFO( "ARTS_P_BUFFER_SIZE      "<<arts_stream_get( _arts_stream, ARTS_P_BUFFER_SIZE ));
    INFO( "ARTS_P_BUFFER_TIME      "<<arts_stream_get( _arts_stream, ARTS_P_BUFFER_TIME ));
    INFO( "ARTS_P_BUFFER_SPACE     "<<arts_stream_get( _arts_stream, ARTS_P_BUFFER_SPACE ));
    INFO( "ARTS_P_SERVER_LATENCY   "<<arts_stream_get( _arts_stream, ARTS_P_SERVER_LATENCY ));
    INFO( "ARTS_P_TOTAL_LATENCY    "<<arts_stream_get( _arts_stream, ARTS_P_TOTAL_LATENCY ));
    INFO( "ARTS_P_BLOCKING         "<<arts_stream_get( _arts_stream, ARTS_P_BLOCKING ));
    INFO( "ARTS_P_PACKET_SIZE      "<<arts_stream_get( _arts_stream, ARTS_P_PACKET_SIZE ));
    INFO( "ARTS_P_PACKET_COUNT     "<<arts_stream_get( _arts_stream, ARTS_P_PACKET_COUNT ));
	INFO( "ARTS_P_PACKET_SETTINGS  "<<arts_stream_get( _arts_stream, ARTS_P_PACKET_SETTINGS ));

    setState( state_open );
    _is_initialized = true;

    return;
}


void 
OutARTSStream::putFrame( Frame & f )
{

    AudioBuffer * a = f.makeABuf();
    if ( !a ) {
        /* this frame has no audio */
        INFO( "no audio in this frame" );
        return;
    }

    if ( a->getNChannels() != _nChannels ) {
        WARN( "audio input nChannels != my nchannels "
              <<PRINTV( a->getNChannels() ) << PRINTV( _nChannels ) );
        return;
    }

    /*   either the channels are in sepearete buffers, or interleaved
     *   in a single buffer. If they are in seperate buffers, we are
     *   going to interleave them into a single buffer
     */
    int16_t * buffer = (int16_t*)a->interleave()->buf();

    /* Send the audio to the device */
    int total = a->getNSamples() * _nChannels * sizeof( int16_t );
    int written = 0;
    while ( written < total) {
        int result = arts_write( _arts_stream, buffer + written, total - written);
        if ( result <= 0 ) {
            WARN( "error writing to arts " << arts_error_text( result ) );
            return;
        }
        written += result;
    }


    return;

}

PropertyNode * 
OutARTSStream::getProperties( PropertyNode * p ) const
{
    WARN( "FIXME" );
    return 0;
}
