/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, Rolf.Dubitzky@gmx.net
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// STDC
#include <unistd.h>
#include <getopt.h>

// STDC++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// libxml
#include <libxml/parser.h>

// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/player.hh"
#include "libpiave/mediaelements.hh"
#include "libpiave/exception.hh"
#include "libpiave/property.hh"
#include "libpiave/plugin.hh"
#include "libpiave/pluginmgr.hh"
#include "libpiave/streammgr.hh"
#include "libpiave/bufferpool.hh"
#include "libpiave/VEML.hh"
#include "libpiave/scenebuilder.hh"

using namespace PIAVE;

StoryBoard * openVEMLFile( string filename );
StoryBoard * openVideoFile( string filename );

void usage( int argc,  char * const * argv  ) { 
    cout << "usage:  "<<argv[0]<<" [options] <infile> <outfile>"<<endl;
    cout << " -v <level>       level of verbosity (default: 0=quiet) " << endl;
    cout << " -d <dir>         search <dir> recursively for plugins" << endl;
    cout << " -l <lib>         load plugin library <lib>" << endl;
    exit( 1 ); 
}

int
main( int argc,  char * const * argv  )
{
    bool   default_plugin_src   = true;
    int    n_frames             = -1;
    int    n_plugins            = 0;
    double scale                = 1.0;
    double speed                = 1.0;

    int c;
    opterr = 0;
    while ((c = getopt (argc, argv, "d:l:f:A:V:s:S:v:h")) != -1) {
        switch (c)
        {
          case 'v':
              PIAVE::Global::verbosity = atoi( optarg );
              break;
          case 'h':
              usage( argc, argv );
              break;
          case 'd':
              n_plugins += PluginMgr::openDLLsFromDir( optarg );
              default_plugin_src = false;
              break;
          case 'l':
              n_plugins += PluginMgr::openDLL( optarg ); 
              default_plugin_src = false;
              break;
          case 'f':
              n_frames = atoi( optarg );
              break;
          default:
              cout <<"unknown option: "<<(char)c<<endl;
              usage( argc, argv );
              break;
        }
    }

    if ( optind != argc-2 ) {
        cout << "Missing filename" << endl;
        usage( argc, argv );
    }

    string infilename( argv[optind] );
    string outfilename( argv[optind+1] );

    INFO( PRINTV( infilename ) );
    INFO( PRINTV( outfilename ) );

    /*
     *    init PIAVE 
     */
    PIAVE::Init();
    PluginMgr::openDLLsFromDefaultPath();


    /*
     *    try to open <file> as a VEML or Video file
     */
    StoryBoard * story = openVEMLFile( infilename );
    if ( !story ) {
        story = openVideoFile( infilename );
    }
    if ( !story ) {
        cout << "ERROR : couldn't open file \""<<infilename<<"\"" << endl;
        return 1;
    }
    
    /*
     *  
     */
    OutAVStreamIFace * voutstream = 0;
    PropertyNode * file_caps = new PropertyNode( "outstream" );
    PropertyNode * file = file_caps->addDaughter( "file" );
    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "rawdv" );
    Plugin * p = PluginMgr::find( file_caps );
    if ( !p ) {
        cout << "WARNING: no output plugin to handle file output" << endl;
    } else {
        PropertyNode * file_props = new PropertyNode( "props" );
        file_props->addAttr( "filename", outfilename );
        PluginBase * b = p->newElement( file_props );
        if ( !b ) {
            cout << "WARNING : instantiating plugin "<<p->getName()<<endl;
        } else {
            voutstream = static_cast<OutAVStreamIFace*>( b );
            if ( voutstream==0 ) {
                cout << "WARNING : wrong plugin type " << endl;
            }
        }
        PDELETE( file_props );
    }
    PDELETE( file_caps );


    /*
     *    main loop
     */
    Time pos = 0;
    int nProcessedFrames = 0;
    while ( true ) {
        /* stop ? */
        if ( nProcessedFrames == n_frames ) break;
        if ( pos >= story->length() ) break;

        /* go */
        Frame f = story->getFrame( pos );
        voutstream->putFrameSave( f );
        pos += Global::renderFmt.getFrameLength();
        nProcessedFrames++;
    }


    /*
     *    print some stats
     */
    BufferPool::printStats();
    story->printStats();

    PDELETE( voutstream );
    PDELETE( story );

    return 0;
}


StoryBoard *
openVEMLFile( string filename )
{
    /*
     *  all functionality here must be moved to more proper locations
     */
    INFO( "trying to open VEML file" );

    xmlDocPtr cmndDoc = xmlParseFile( filename.c_str() );
    if ( !cmndDoc ) {
        INFO( "couldn't parse command ... no XML ?" );
        return 0;
    }

    xmlNodePtr messageRoot = xmlDocGetRootElement( cmndDoc );
    if ( !messageRoot ) {
        xmlFreeDoc( cmndDoc );
        INFO( "no root node ... no XML?" );
        return 0;
    }

    StorySeqNode * scenes = 0;
    for ( xmlNodePtr xmlCmndNode = messageRoot->xmlChildrenNode;
          xmlCmndNode != 0;
          xmlCmndNode = xmlCmndNode->next )
    {
        if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"scenelist") ) {
            VEMLCmnd * cmnd = new VEMLCmnd( "scenelist" );
            cmnd->copyFromXMLNode( cmndDoc, xmlCmndNode );
	    INFO( cmnd->toXML() );
            scenes = SceneBuilder::buildScenelist( cmnd );
            delete cmnd;
        } else if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"outstream") ) {
            /* parse outstream properties */
            WARN( "FIXME" );
        } else if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"head") ) {
            /* FIXME move t a better place */
            /*
             *   <head>
             *     <path>/some/absolute/path/to/files</path>
             *     <path>/some/other/absolute/path/to/files</path>
             *   </head>
             *   StreamMgr::addPathEntry( );
             */
        }
    }
    
    if ( !scenes ) return 0;

    xmlFreeDoc( cmndDoc );
 
    /*
     *    setup render tree
     */
    StoryBoard * story = new StoryBoard( "MasterBoard" );
    story->setHead( scenes );
//    story->setProperties( props );

    return story;
}


StoryBoard * 
openVideoFile( string filename )
{
    InAVFStream * instream = 0;
    try {
        instream = StreamMgr::getAVInput( filename );
    } catch ( const file_error & e ) {
        cerr << "caught PIAVE file error: "<<e.what() << endl;
        return 0;
    } catch ( const std::exception & e ) {
        cerr << "caught std exception: "<<e.what() << endl;
        return 0;
    }
    
    if ( !instream ) {
        cerr << " Couldn't open file "<<filename<<endl;
        return 0;
    }

    /*
     *    setup render tree
     */
    AVChunk * avChunk = new AVChunk( instream );
    StoryBoard * story = new StoryBoard( "MasterBoard" );
    story->setHead( avChunk );
//    story->setProperties( props );

    return story;
}
